# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
A composite that scales problem variables as directed. If a scaling value is
not specified, calculates it based on quadratic and bias ranges.

"""

try:
    from dwave.preprocessing import ScaleComposite as _ScaleComposite
except ImportError:
    from dimod.reference.composites._preprocessing import NotFound as _ScaleComposite

from dimod.reference.composites._preprocessing import DeprecatedToPreprocessing


__all__ = ['ScaleComposite']


class ScaleComposite(DeprecatedToPreprocessing, _ScaleComposite):
    pass


ScaleComposite.__doc__ = _ScaleComposite.__doc__
